import subprocess
from pathlib import Path
from tempfile import NamedTemporaryFile
from textwrap import dedent
from typing import Optional, Union

from cmstp.utils.command import CommandKind
from cmstp.utils.common import (
    PACKAGE_BASH_HELPERS_PATH,
    PIPX_PYTHON_PATH,
    FilePath,
)


def run_script_function(
    script: FilePath,
    function: Optional[str] = None,
    run: bool = True,
    capture_output: bool = False,
) -> Union[str, subprocess.CompletedProcess]:
    """
    Build a wrapper script string for the specified command kind and possibly execute it.
    Args:
        script (Path | str):      The path to the script to source or execute.
        function (Optional[str]): The function within the script to call. If None, the script is executed directly.
        run (bool):               If True, executes the script. Otherwise, returns the string.
        capture_output (bool):    If True, captures the output of the script. Ignored if run=False.
    Returns:
        str | subprocess.CompletedProcess: The generated script string (if run=False)
                                           or the subprocess result (if run=True).
    """
    if not Path(script).exists():
        raise FileNotFoundError(f"Script file not found: {script}")
    # TODO: Add check for function existence. Make util for this somewhere and use both here and in scheduler

    kind = CommandKind.from_script(script)
    if kind == CommandKind.BASH:
        return _run_bash_script_function(script, function, run, capture_output)
    elif kind == CommandKind.PYTHON:
        return _run_python_script_function(
            script, function, run, capture_output
        )
    else:
        raise ValueError(
            f"Unsupported script type: {kind} (supported: {CommandKind.BASH.name}, {CommandKind.PYTHON.name})"
        )


def _run_bash_script_function(
    script: FilePath,
    function: Optional[str] = None,
    run: bool = True,
    capture_output: bool = False,
) -> Union[str, subprocess.CompletedProcess]:
    """
    Build a bash wrapper script string and possibly execute it.
    Args:
        script (Path | str):      The path to the bash script to source or execute.
        function (Optional[str]): The function within the script to call. If None, the script is executed directly.
        run (bool):               If True, executes the script. Otherwise, returns the string.
        capture_output (bool):    If True, captures the output of the script. Ignored if run=False.
    Returns:
        str | subprocess.CompletedProcess: The generated bash script string (if run=False)
                                           or the subprocess result (if run=True).
    """
    # Source pipx venv and helpers
    sourcing = dedent(
        f"""\
        source {PIPX_PYTHON_PATH.parent / 'activate'}
        source {PACKAGE_BASH_HELPERS_PATH}
    """
    )

    # Build script body
    if function:
        # Simply source and call function
        body = sourcing + dedent(
            f"""\
            source {script}
            {function} "$@"
        """
        )
    else:
        # Create temporary sourcing file for usage with BASH_ENV
        sourcing_file = NamedTemporaryFile(
            mode="w", suffix=".bash", prefix="sourcing_", delete=False
        )
        sourcing_file.write(sourcing)
        sourcing_file.flush()
        sourcing_file.close()

        # Run the script with BASH_ENV set
        body = dedent(
            f"""\
            export BASH_ENV='{sourcing_file.name}'
            {CommandKind.BASH.exe} {script} "$@"
        """
        )

    # (Run) Full bash script
    wrapper_src = (
        dedent(
            """\
        #!/usr/bin/env bash
        set -euo pipefail
    """
        )
        + body
    )
    if run:
        return subprocess.run(
            [CommandKind.BASH.exe, "-c", wrapper_src],
            capture_output=capture_output,
            text=True,
        )

    return wrapper_src


def _run_python_script_function(
    script: FilePath,
    function: Optional[str] = None,
    run: bool = True,
    capture_output: bool = False,
) -> Union[str, subprocess.CompletedProcess]:
    """
    Build a Python wrapper script string and possibly execute it.
    Args:
        script (Path | str):      The path to the Python script to import or execute.
        function (Optional[str]): The function within the script to call. If None, the script is executed directly.
        run (bool):               If True, executes the script. Otherwise, returns the string.
        capture_output (bool):    If True, captures the output of the script. Ignored if run=False.
    Returns:
        str | subprocess.CompletedProcess: The generated Python script string (if run=False)
                                           or the subprocess result (if run=True).
    """
    if function:
        # Import the module dynamically and call the function
        wrapper_src = dedent(
            f"""\
            import importlib.util, sys
            p = {repr(str(script))}
            spec = importlib.util.spec_from_file_location('_run_mod', p)
            mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(mod)
            func = getattr(mod, {repr(function)})
            res = func(*sys.argv[1:])
            if isinstance(res, int):
                sys.exit(res)
        """
        )
    else:
        # Just execute the script directly
        wrapper_src = dedent(
            f"""\
            import sys
            from pathlib import Path
            script = Path({repr(str(script))})
            sys.path.insert(0, str(script.parent))
            with open(script, 'rb') as f:
                code = compile(f.read(), script, 'exec')
                exec(code, {{'__name__': '__main__'}})
        """
        )

    if run:
        return subprocess.run(
            [CommandKind.PYTHON.exe, "-c", wrapper_src],
            capture_output=capture_output,
            text=True,
        )

    return wrapper_src


def bash_check(check_name: str) -> subprocess.CompletedProcess:
    """Run a (helper) check function"""
    # Create a mock bash file (used only as a placeholder)
    with NamedTemporaryFile(
        mode="w", suffix=".bash", delete=False
    ) as tmp_file:
        tmp_file.write("#!/usr/bin/env bash\n")
        tmp_file_path = Path(tmp_file.name)

    try:
        # Try running the helper function
        return run_script_function(
            tmp_file_path, check_name, run=True, capture_output=True
        )
    except Exception as e:
        # Return a failed CompletedProcess instead of None
        return subprocess.CompletedProcess(
            args=[str(tmp_file_path), check_name],
            returncode=1,
            stdout=b"",
            stderr=str(e).encode(),
        )
    finally:
        # Always clean up
        tmp_file_path.unlink(missing_ok=True)



import argparse
import json
import sys
from pathlib import Path
from typing import List, Tuple

from cmstp.core.logger import Logger, LoggerSeverity
from cmstp.utils.system_info import SystemInfo


def get_config_args(
    args=sys.argv[1:],
) -> Tuple[SystemInfo, Path, List[str]]:
    """Parse command-line arguments and return system info, config info, and remaining args."""
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("--system-info", type=str, default=None)
    # parser.add_argument("--simulate-hardware", action="store_true")  # Now included in system info
    # parser.add_argument(
    #     "--config-directory",
    #     type=Path,
    #     default=None
    # )
    parser.add_argument("--config-file", type=Path, default=None)
    args, remaining = parser.parse_known_args(args)

    # System info
    system_info = {}
    if args.system_info:
        try:
            # Parse JSON input
            system_info = json.loads(args.system_info)
            if not isinstance(system_info, dict):
                raise ValueError("The value for --system-info must be a dict.")
        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid JSON format for --system-info: {e}")

    # Config file
    if args.config_file is not None and not args.config_file.exists():
        Logger.logrichprint(
            LoggerSeverity.FATAL, f"Config file not found: {args.config_file}"
        )
        raise FileNotFoundError

    return system_info, args.config_file, remaining


# Example function usage
def _example_function(*args):
    system_info, config_info, remaining = get_config_args(args)

    # Use argparse for the rest
    parser = argparse.ArgumentParser()
    parser.add_argument("--foo")
    parser.add_argument("--bar")
    remaining_args = parser.parse_args(remaining)

    foo_arg = remaining_args.foo


# Example main usage
if __name__ == "__main__":
    system_info, config_info, remaining = get_config_args()

    # Use argparse for the rest
    parser = argparse.ArgumentParser()
    parser.add_argument("--foo")
    parser.add_argument("--bar")
    remaining_args = parser.parse_args(remaining)

    foo_arg = remaining_args.foo



import os
import tempfile

from ruamel.yaml import YAML

from cmstp.utils.tasks import HARDWARE_SPECIFIC_TASKS

# from cmstp.main import main


def test_hardware_specific():
    """Test enabling only the hardware-specific tasks in a temporary YAML file"""

    # Enable all hardware-specific tasks in a temporary YAML file
    data = {task: {"enabled": True} for task in HARDWARE_SPECIFIC_TASKS}

    # Create the temporary YAML file
    yaml = YAML()
    yaml.indent(mapping=2, sequence=4, offset=2)
    yaml.default_flow_style = False
    with tempfile.NamedTemporaryFile(
        mode="w", suffix=".yaml", delete=False
    ) as tmpfile:
        yaml.dump(data, tmpfile)
        temp_yaml = tmpfile.name

    # # Run main() with this temporary YAML file
    # main(["--config-file", temp_yaml])

    # Delete the temporary file
    os.remove(temp_yaml)



import inspect
import platform
import subprocess
from typing import Optional, TypedDict

import distro


class SystemInfo(TypedDict):
    # fmt: off
    type:              str
    kernel:            str
    name:              Optional[str]
    codename:          Optional[str]
    version_dot:       Optional[str]
    version_nodot:     Optional[str]
    arch:              str
    simulate_hardware: bool
    manufacturer:      str
    # fmt: on


def get_architecture() -> str:
    """Retrieve the system architecture using dpkg."""
    result = subprocess.run(
        ["dpkg", "--print-architecture"],
        text=True,
        capture_output=True,
    )
    if result.returncode != 0:
        raise RuntimeError(
            f"Failed to retrieve architecture info via dpkg: {result.stderr.strip()}"
        )
    else:
        return result.stdout.strip().lower()


# TODO: Test if this works without having to install "dmidecode" first (via apt) on a fresh system
def get_manufacturer() -> str:
    """Retrieve the system manufacturer using dmidecode."""
    result = subprocess.run(
        ["sudo", "dmidecode", "-s", "system-manufacturer"],
        text=True,
        capture_output=True,
    )
    if result.returncode != 0:
        raise RuntimeError(
            f"Failed to retrieve manufacturer info via dmidecode: {result.stderr.strip()}"
        )
    else:
        return result.stdout.strip().lower()


def get_system_info() -> SystemInfo:
    """Retrieve detailed information about the host operating system."""

    def is_current_test(test_name: str) -> bool:
        """
        Returns True if the current call stack indicates that we're
        running inside the given pytest test function name.
        """
        for frame_info in inspect.stack():
            func = frame_info.function
            # Typical pytest test functions start with "test_"
            if func.startswith("test_"):
                return func == test_name
        return False

    system_info = SystemInfo()
    # linux, darwin, etc.
    system_info["type"] = platform.system().lower()
    # x86_64, aarch64, etc.
    system_info["kernel"] = platform.machine()
    # Simulate Hardware (e.g. GPU)
    system_info["simulate_hardware"] = is_current_test(
        "test_hardware_specific"
    )
    if system_info["type"] == "linux":
        # ubuntu, debian, etc.
        system_info["name"] = distro.id().lower()
        # focal, jammy, buster, bullseye, etc.
        system_info["codename"] = distro.codename().lower()
        # 20.04, 22.04, etc.
        system_info["version_dot"] = distro.version()
        # 2004, 2204, etc.
        system_info["version_nodot"] = distro.version().replace(".", "")
        # amd64, arm64, etc.
        system_info["arch"] = get_architecture()
        # manufacturer
        system_info["manufacturer"] = get_manufacturer()
    else:
        # Unsupported OS
        system_info["name"] = None
        system_info["codename"] = None
        system_info["version_dot"] = None
        system_info["version_nodot"] = None
        system_info["arch"] = None
        system_info["manufacturer"] = None

    return system_info



import grp
import os
import pwd
import shutil
import sys
from importlib import resources
from pathlib import Path
from tempfile import mkdtemp, mkstemp
from typing import Optional, Union

from cmstp.utils.patterns import PatternCollection

PACKAGE_SRC_PATH = Path(resources.files("cmstp")).expanduser().resolve()
PACKAGE_CONFIG_PATH = PACKAGE_SRC_PATH / "config"
PACKAGE_TESTS_PATH = PACKAGE_SRC_PATH.parents[1] / "tests"
PACKAGE_BASH_HELPERS_PATH = (
    PACKAGE_SRC_PATH / "scripts" / "bash" / "helpers" / "helpers.bash"
)
PIPX_PYTHON_PATH = Path(sys.executable)

FilePath = Union[Path, str]


def generate_random_path(
    suffix: Optional[str] = None,
    prefix: Optional[str] = None,
    create: bool = False,
) -> Path:
    if suffix is not None and suffix.startswith("."):
        # File
        fd, path = mkstemp(suffix, prefix)
        os.close(fd)
        if not create:
            os.remove(path)
    else:
        # Directory
        path = mkdtemp(suffix, prefix)
        if not create:
            shutil.rmtree(path)

    return Path(path)


def is_simple_path(path: FilePath) -> bool:
    invalid = '<>:"|?*@'
    return not any(ch in invalid for ch in str(path))


def resolve_package_path(raw_script: FilePath) -> Optional[FilePath]:
    """
    Resolve paths that may refer to package resources. Package paths are in the format:
        "package://<package-name>/relative/path/inside/package"
    The input type (should be Path or str) is preserved in the output.
    """
    # Return wrong types as-is
    if not isinstance(raw_script, (Path, str)):
        return raw_script

    # Resolve package paths
    match = PatternCollection.PATH.patterns["package"].match(str(raw_script))
    if match:
        pkg_name, rel_path = match.groups()
        try:
            resolved_path = Path(resources.files(pkg_name)) / rel_path
        except ModuleNotFoundError:
            return None
    else:
        # NOTE: We use 'os' and no built-in 'Path' method to retain '<type>://' multiple slashes
        resolved_path = os.path.expanduser(str(raw_script))

    # Return same type as input
    if isinstance(raw_script, Path):
        return Path(resolved_path)
    else:  # str
        return str(resolved_path)


def revert_sudo_permission(path: FilePath) -> None:
    # Determine the real user who invoked sudo
    user = os.environ.get("SUDO_USER") or os.environ["USER"]
    uid = pwd.getpwnam(user).pw_uid
    gid = grp.getgrnam(user).gr_gid

    for root, dirs, files in os.walk(path):
        for d in dirs:
            full = os.path.join(root, d)
            os.chown(full, uid, gid)
            os.chmod(full, 0o775)  # drwxrwxr-x

        for f in files:
            full = os.path.join(root, f)
            os.chown(full, uid, gid)
            os.chmod(full, 0o664)  # -rw-rw-r--

    # Also fix the top-level directory
    os.chown(path, uid, gid)
    os.chmod(path, 0o775)



from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, Optional

from ruamel.yaml import YAML

from cmstp.utils.common import resolve_package_path


def load_yaml(yaml_file: Path) -> Optional[Dict[str, Any]]:
    """Load a YAML file and normalize its content."""

    def join_constructor(loader, node):
        """Custom !join constructor to concatenate strings and aliases."""
        parts = loader.construct_sequence(node)
        return "".join(parts)

    def normalize_yaml(obj):
        """
        Recursively normalize YAML content:
        - Convert all numbers to float
        - Remove duplicates in lists
        """
        if not obj:
            # Empty or None
            return obj

        if isinstance(obj, dict):
            # Recurse
            return {k: normalize_yaml(v) for k, v in obj.items()}

        elif isinstance(obj, list):
            # Remove duplicates
            unique_list = []
            for item in obj:
                if item not in unique_list:
                    unique_list.append(item)

            # Normalize items (float/int)
            return [normalize_yaml(item) for item in unique_list]

        elif isinstance(obj, str):
            # Resolve package paths
            return resolve_package_path(obj)

        elif isinstance(obj, bool):
            # Keep booleans as-is
            return obj

        # ATTENTION: bool would evaluate as int here
        elif isinstance(obj, (int, float)):
            # Convert all numbers to float
            return float(obj)

        else:
            # Should not happen
            pass

    if not yaml_file.exists():
        return None

    yaml = YAML(typ="safe")
    yaml.constructor.add_constructor("!join", join_constructor)
    with open(yaml_file, "r") as f:
        try:
            content = yaml.load(f) or {}
        except Exception:
            return None
    return normalize_yaml(content)


def overlay_dicts(
    base: Dict, overlay: Dict, allow_default: bool = False
) -> Dict:
    """
    Recursively overlay overlay-dict onto base-dict.
    Keys in overlay replace or update those in base, unless the value is "default".
    """
    overlayed_dict = deepcopy(base)
    for key, value in overlay.items():
        if allow_default and key in overlayed_dict and value == "default":
            # Keep base value
            continue
        elif (
            key in overlayed_dict
            and isinstance(overlayed_dict[key], dict)
            and isinstance(value, dict)
        ):
            # Recursively overlay nested dicts
            overlayed_dict[key] = overlay_dicts(
                overlayed_dict[key], value, allow_default
            )
        else:
            # Directly set/replace value
            overlayed_dict[key] = value
    return overlayed_dict



import subprocess
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional, Tuple

from cmstp.core.logger import Logger
from cmstp.utils.common import generate_random_path, resolve_package_path
from cmstp.utils.git_repos import clone_git_files, is_git_repo
from cmstp.utils.system_info import get_system_info


@dataclass
class MainSetupArgs:
    # fmt: off
    tasks:               List[str] = field(init=False, default_factory=list)
    config_file:         Path      = field(init=False, default=None)
    config_directory:    Path      = field(init=False, default=None)
    disable_preparation: bool      = field(init=False, default=False)
    # fmt: on


@dataclass
class MainSetupProcessor:
    # fmt: off
    logger: Logger        = field(repr=False)
    args:   MainSetupArgs = field(repr=False)
    argv:   List[str]     = field(repr=False)
    # fmt: on

    def process_args(self) -> Tuple[MainSetupArgs, Optional[Path]]:
        main_setup_args = MainSetupArgs()
        cloned_config_dir = None

        # Tasks
        main_setup_args.tasks = self.args.tasks or []

        # Config directory
        if is_git_repo(str(self.args.config_directory)):
            # Git repo
            cloned_config_dir = generate_random_path(
                prefix="cmstp_config_dir_"
            )
            cloned_path = clone_git_files(
                str(self.args.config_directory), dest_path=cloned_config_dir
            )
            if cloned_path is None:
                self.logger.fatal(
                    f"Failed to clone config directory "
                    f"git repo '{self.args.config_directory}'",
                )
            self.args.config_directory = cloned_path
        else:
            # Local path
            self.args.config_directory = resolve_package_path(
                self.args.config_directory
            )
            if (
                self.args.config_directory is not None
                and self.args.config_directory.exists()
            ):
                if self.args.config_directory.is_file():
                    # If a file is specified, use its parent directory
                    self.args.config_directory = (
                        self.args.config_directory.parent
                    )
                # else: It's a directory, use as is
            else:
                self.logger.fatal(
                    f"Config directory not found: {self.args.config_directory}",
                )
        main_setup_args.config_directory = self.args.config_directory

        # Config file
        if self.args.tasks and not any(
            arg in self.argv for arg in ("-f", "--config-file")
        ):
            # If tasks are specified without a config file, ignore the config file
            self.args.config_file = None
        elif not self.args.config_file.exists():
            # If a config directory is specified, look for a config file there
            possible_config_file = (
                self.args.config_directory / self.args.config_file
            )
            if possible_config_file.exists():
                self.args.config_file = possible_config_file
            else:
                self.logger.fatal(
                    f"Config file not found: {self.args.config_file}",
                )
        main_setup_args.config_file = self.args.config_file

        # Disable preparation
        main_setup_args.disable_preparation = self.args.disable_preparation

        self.logger.debug(
            f"Processed main setup args: {repr(main_setup_args)}"
        )

        return main_setup_args, cloned_config_dir

    def check_system_compatibility(self) -> None:
        # Check system information
        system_info = get_system_info()
        if system_info["name"] is None:
            self.logger.fatal(
                f"Unsupported OS type '{system_info['type']}'",
            )

        self.logger.debug(f"System information: {system_info}")

    def prepare(self) -> None:
        error_msg = None
        requirements_id = self.logger.add_task("install-requirements", total=2)

        result_update = subprocess.run(
            ["sudo", "apt-get", "update"],
            capture_output=True,
            text=True,
        )
        self.logger.update_task(requirements_id, "Updated apt packages")

        result_upgrade = subprocess.run(
            ["sudo", "apt-get", "-y", "upgrade"],
            capture_output=True,
            text=True,
        )
        self.logger.update_task(requirements_id, "Upgraded apt packages")

        success = (
            result_update.returncode == 0 and result_upgrade.returncode == 0
        )
        self.logger.finish_task(requirements_id, success=success)

        if not success:
            error_msg = "Failed to update/upgrade apt packages"
            if result_update.returncode != 0:
                error_msg += f"\nUpdate output:\n{result_update.stdout}\n{result_update.stderr}"
            if result_upgrade.returncode != 0:
                error_msg += f"\nUpgrade output:\n{result_upgrade.stdout}\n{result_upgrade.stderr}"
            self.logger.fatal(error_msg)

        self.logger.debug("System preparation completed successfully")

        return success
